/* Birth-deah process with local density dependency in two dimensional torus space*/
/* Implemented by FT takasu@ics.nara-wu.ac.jp */
/* 2008 March 19 */
/* Birth rate is B1 - B2*N_{local}
   Death rate is D1 + D2*N_{local}. N_{local} is the local density that an individual feels around it. */
/* New offspring disperses from the parent. */

/* No-parallelization version */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include "omp.h"

#define T_END	100
#define B1		0.4
#define B2		0.0
#define D1		0.2
#define D2		0.001

#define L				1.0		/* Scale of the space */
#define SD_COMPETITION	0.2		/* Length of effective interaction with others */
#define SD_DISPERSAL	0.2		/* Dispersal distance of new born individual */
#define PI				3.141592654

/* Definition of individual's properties */
typedef struct individual INDIVIDUAL;
struct individual {
    double time_to_birth, time_to_death;
	double x, y, local_density;
    struct individual *next_indiv;
};

extern double ran2(long);	/* Uniform random number generator in [0, 1) */
extern double expdev(long); /* Exponential random number generator with mean 1 */
extern double gasdev(long);	/* Gaussian random number generator with mean 0 and variance 1 */

void set_interevent_times(INDIVIDUAL*), set_interevent_times_omp(INDIVIDUAL*), make_an_event_happen(INDIVIDUAL*, double*);
void display_list(INDIVIDUAL*), calc_local_densities(INDIVIDUAL*), calc_local_densities_omp(INDIVIDUAL*), write_to_file(FILE*, INDIVIDUAL*);
INDIVIDUAL *make_new_indiv(), *add_next_indiv(INDIVIDUAL*);
double minimum(double, double), weight(double), distance(INDIVIDUAL*, INDIVIDUAL*), torus(double), sqr(double);
int number_of_indivs(INDIVIDUAL*);

long seed;
FILE *fp;

int main(){

	INDIVIDUAL *list_head, *indiv;
	int i;
	double t, dt;
	double t_next = 1.0, t_eq_intv = 0.0;
  	time_t	time1, time2;

	time(&time1);

	fp = fopen("data-SW=0.2-SM=0.2","w");
	
	/* Initialization of random number generator */
	seed = (long)time(NULL);
	ran2(-seed);
  	
	/* Create the root of list */
	list_head = make_new_indiv(); 

	/* We start from ten individuals randomly located in the space */
	indiv = list_head;
	for(i=0; i<10; i++){
		indiv = add_next_indiv(indiv);
		indiv->x = L*ran2(seed);
		indiv->y = L*ran2(seed);
		indiv->local_density = 0.0;
	}

	/* Let's start! */
	t = 0.0;
	while( t < T_END){
    
		/* Report the number of individuals every unit time */
		while( t_eq_intv < t + t_next ){
			if( t_eq_intv > T_END ) break;
				printf("%f, H = %d\n", t_eq_intv, number_of_indivs(list_head));
				write_to_file(fp, list_head);
				t_eq_intv += 1.0;
		}

		/* Calculate local density for all individuals */
		calc_local_densities(list_head);
		
		/* Assign interevent times to all individuals. Now local density does affect interevent times */
		set_interevent_times(list_head); 

		/* An individual gives birth or death and the time advances by dt (This is continuos-time process!) */
		make_an_event_happen(list_head, &dt);
		t += dt;
	
	} /* end of while t < T_END */
	
	time(&time2);
	
	printf("%d %d, %d\n", time1, time2, time2 - time1);

	return 0;
}

void calc_local_densities(INDIVIDUAL *list)
/* This calculates local density for all individuals connected to the "list" */
{
	INDIVIDUAL *indiv1, *indiv2;
	double sum;
	
	for( indiv1 = list->next_indiv; indiv1 != NULL; indiv1 = indiv1->next_indiv ){
	
		sum = 0.0;
		for( indiv2 = list->next_indiv; indiv2 != NULL; indiv2 = indiv2->next_indiv )
			sum += weight( distance(indiv1, indiv2) );
		sum -= weight( 0.0 ); /* Exclude the self */
		
		indiv1->local_density = sum;
		
	}
}

double weight(double x)
/* Competition kernel. Weighting factor to calculate local density */
{
	double tmp;
	
	//tmp = exp( -sqr(x)/2/sqr(SD_COMPETITION) )/2.0/PI/sqr(SD_COMPETITION);
	tmp = exp( -sqr(x)/2/sqr(SD_COMPETITION) );
	
	return tmp;
}

double distance(INDIVIDUAL *indiv1, INDIVIDUAL *indiv2)
/* Calculate the distance between two individuals. But remember our space is torus. */
{
	double tmpx, tmpy, tmpLx, tmpLy, distx, disty;
	
	tmpx = fabs(indiv1->x - indiv2->x);
	tmpy = fabs(indiv1->y - indiv2->y);
	tmpLx = L - tmpx;
	tmpLy = L - tmpy;
	
	if(tmpx < tmpLx)
		distx = tmpx;
	else
		distx = tmpLx;
		
	if(tmpy < tmpLy)
		disty = tmpy;
	else
		disty = tmpLy;

	return (sqrt( distx*distx + disty*disty) );
}

double torus(double x)
{
	double pos_x, diff;
	int int_x;
	
	pos_x = x;
	
	/* All individuals shall be landed within the space 0 < x, y < L */
	if( pos_x >= 0){
		int_x = 0;
		while( diff = pos_x - int_x*L > 0){
			int_x++;
		}
		pos_x = pos_x - (int_x-1)*L;
	}

	if( pos_x < 0){
		int_x = 0;
		while( diff = pos_x + int_x*L < 0){
			int_x++;
		}
		pos_x = pos_x + int_x*L;
	}

	return pos_x;
}


void set_interevent_times(INDIVIDUAL *list)
/* Assign interevent times to all individuals */
{
	INDIVIDUAL *indiv;
	double time_to_birth, time_to_death;
	int pop_size;
  
	/* Scan all individuals to assign interevent times */
	for(indiv = list->next_indiv; indiv != NULL; indiv = indiv->next_indiv){
		
		time_to_birth = expdev(seed)/(B1 - B2*indiv->local_density); /* Interevent time for birth is exponentially distributed */
		time_to_death = expdev(seed)/(D1 + D2*indiv->local_density); /* Interevent time for death is exponentially distributed */
	
		indiv->time_to_birth = time_to_birth;
		indiv->time_to_death = time_to_death;
	}

}

void make_an_event_happen(INDIVIDUAL *list, double *dt)
/* Birth or death happens to an individual and the time proceeds by dt */
{
	INDIVIDUAL *indiv, *prev, *target, *target_prev, *target_next, *new_born;
	double min_time, min_time_test, dist, angle;
	
	/* Start from the first individual in the list */
	target = list->next_indiv;
	target_prev = list;

	/* If there is at least one individual ... */
	if( target != NULL ){
	
		/* minimum(x, y) returns a smaller value */
		min_time = minimum(target->time_to_birth, target->time_to_death); 
	
		/* Look for an individual who has the smallest interevent time to birth or death */
		for(prev = list, indiv = list->next_indiv; indiv != NULL; prev = prev->next_indiv, indiv = indiv->next_indiv){

			min_time_test = minimum( indiv->time_to_birth, indiv->time_to_death );
		
			if( min_time_test < min_time ){
				min_time = min_time_test;
				target = indiv;
				target_prev = prev;
			}
      			
		} /* end of for prev, indiv */

		if( target->time_to_birth < target->time_to_death ){
			/* target gives birth to an individual */
			new_born = make_new_indiv();
			
			dist = SD_DISPERSAL*gasdev(seed);
			angle = 2*PI*ran2(seed);
			new_born->x = torus( target->x + dist*cos(angle) );
			new_born->y = torus( target->y + dist*sin(angle) );
			
			target_next = target->next_indiv;
			target->next_indiv = new_born;
			new_born->next_indiv = target_next;
		} else {
			/* target dies*/
			target_prev->next_indiv = target->next_indiv;
			free(target);
		}
    
		*dt = min_time;
    	
	} else *dt = 1.0;
  
}

INDIVIDUAL *make_new_indiv()
/* Create an individual */
{
  INDIVIDUAL *indiv;

  indiv = (INDIVIDUAL *)malloc(sizeof(INDIVIDUAL));
  indiv->time_to_birth = 0.0;
  indiv->time_to_death = 0.0;
  indiv->next_indiv = NULL;
  
  return indiv;
}

INDIVIDUAL *add_next_indiv(INDIVIDUAL *indiv)
/* Create and append an individual AFTER indiv and returns the pointer to the added individual */
{
  INDIVIDUAL *new_indiv;

  if(indiv->next_indiv != NULL)  return NULL;

  new_indiv = make_new_indiv();
  indiv->next_indiv = new_indiv;

  return new_indiv;
}

int number_of_indivs(INDIVIDUAL *list_head)
/* Count the number of individuals */
{
	int count = 0;
	INDIVIDUAL *indiv;
	
	for(indiv = list_head->next_indiv; indiv != NULL; indiv = indiv->next_indiv)
		count++;
	
	return count;
}

void display_list(INDIVIDUAL *list_head)
{
	int count = 0;
	INDIVIDUAL *indiv;

	printf("The list is \n");
	for(indiv = list_head->next_indiv; indiv != NULL; indiv = indiv->next_indiv){
		count++;
		printf("(%f %f %f) ", indiv->x, indiv->y, indiv->local_density);
	}
	printf("The number of individuals is %d \n\n", count);
}

double minimum(double x, double y)
{
	double tmp;
	
	if(x < y)
		tmp = x;
	else
		tmp = y;
	return tmp;
}

double sqr(double x)
{
	return x*x;
}

void write_to_file(FILE *fp, INDIVIDUAL *list)
{
	INDIVIDUAL *indiv;
	
	for(indiv = list->next_indiv; indiv != NULL; indiv = indiv->next_indiv)
		fprintf(fp, "%f %f %f ", indiv->x, indiv->y, indiv->local_density);
	fprintf(fp, "\n");
}
